@@ -74,7 +74,10 @@ test -e $O/include2/asm/atomic.h ||
 test -e $KDIR/include/asm-generic/atomic.h ||
 exit 1
 
-if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
+if
+	grep_q "^VERSION *= *3" $KDIR/Makefile ||
+	grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile
+then
   # do we have gfp_t?
   if grep_q "typedef.*gfp_t" $KDIR/include/linux/gfp.h $KDIR/include/linux/types.h; then
     have_gfp_t=1
@@ -168,16 +171,36 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
   else
     have_netlink_skb_parms=0
   fi
+  if grep_q "eff_cap" $KDIR/include/linux/netlink.h ; then
+    have_netlink_skb_parms_eff_cap=1
+  else
+    have_netlink_skb_parms_eff_cap=0
+  fi
   if grep_q "blk_queue_max_hw_sectors" $KDIR/include/linux/blkdev.h ; then
     need_blk_queue_max_hw_sectors=0
   else
     need_blk_queue_max_hw_sectors=1
   fi
+  if grep_q "blk_queue_max_sectors" $KDIR/include/linux/blkdev.h ; then
+    use_blk_queue_max_sectors_anyways=1
+  else
+    use_blk_queue_max_sectors_anyways=0
+  fi
   if grep_q "blk_queue_max_segments" $KDIR/include/linux/blkdev.h ; then
     need_blk_queue_max_segments=0
   else
     need_blk_queue_max_segments=1
   fi
+  if grep_q "blkdev_get_by_path" $KDIR/include/linux/fs.h ; then
+    have_blkdev_get_by_path=1
+  else
+    have_blkdev_get_by_path=0
+  fi
+  if grep_q "open_bdev_exclusive" $KDIR/include/linux/fs.h ; then
+    have_open_bdev_exclusive=1
+  else
+    have_open_bdev_exclusive=0
+  fi
   if grep_q "typedef.*bool" $KDIR/include/linux/types.h ; then
     have_bool_type=1
   else
@@ -188,6 +211,16 @@ if grep_q "^PATCHLEVEL *= *6" $KDIR/Makefile ; then
   else
     have_sched_timeout_interr=0
   fi
+  if grep_q "fmode_t" $KDIR/include/linux/types.h ; then
+    have_fmode_t=1
+  else
+    have_fmode_t=0
+  fi
+  if grep_q "find_next_zero_bit_le" $KDIR/include/asm-generic/bitops/le.h ; then
+    have_find_next_zero_bit_le=1
+  else
+    have_find_next_zero_bit_le=0
+  fi
 else
     # not a 2.6. kernel. just leave it alone...
     exit 0
@@ -229,16 +262,28 @@ perl -pe "
   { ( $have_set_cpus_allowed_ptr ? '' : '//' ) . \$1}e;
  s{.*(#define KERNEL_HAS_CN_SKB_PARMS.*)}
   { ( $have_netlink_skb_parms ? '' : '//' ) . \$1}e;
+ s{.*(#define HAVE_NL_SKB_EFF_CAP.*)}
+  { ( $have_netlink_skb_parms_eff_cap ? '' : '//' ) . \$1}e;
  s{.*(#define NEED_BLK_QUEUE_MAX_HW_SECTORS.*)}
   { ( $need_blk_queue_max_hw_sectors ? '' : '//' ) . \$1}e;
+ s{.*(#define USE_BLK_QUEUE_MAX_SECTORS_ANYWAYS.*)}
+  { ( $use_blk_queue_max_sectors_anyways ? '' : '//' ) . \$1}e;
  s{.*(#define NEED_BLK_QUEUE_MAX_SEGMENTS.*)}
   { ( $need_blk_queue_max_segments ? '' : '//' ) . \$1}e;
+ s{.*(#define COMPAT_HAVE_BLKDEV_GET_BY_PATH.*)}
+  { ( $have_blkdev_get_by_path ? '' : '//' ) . \$1}e;
+ s{.*(#define COMPAT_HAVE_OPEN_BDEV_EXCLUSIVE.*)}
+  { ( $have_open_bdev_exclusive ? '' : '//' ) . \$1}e;
  s{.*(#define NEED_ATOMIC_ADD_UNLESS.*)}
   { ( $have_atomic_add_unless ? '//' : '' ) . \$1}e;
  s{.*(#define NEED_BOOL_TYPE.*)}
   { ( $have_bool_type ? '//' : '' ) . \$1}e;
  s{.*(#define NEED_SCHEDULE_TIMEOUT_INTERR.*)}
   { ( $have_sched_timeout_interr ? '//' : '' ) . \$1}e;
+ s{.*(#define COMPAT_HAVE_FMODE_T.*)}
+  { ( $have_fmode_t ? '' : '//' ) . \$1}e;
+ s{.*(#define COMPAT_HAVE_FIND_NEXT_ZERO_BIT_LE.*)}
+  { ( $have_find_next_zero_bit_le ? '' : '//' ) . \$1}e;
  " \
 	  < ./linux/drbd_config.h \
 	  > ./linux/drbd_config.h.new
