@@ -257,6 +257,17 @@ void range_check(const enum range_checks what, const char *name,
 		m_strtoll_range(value, 'k', name, DRBD_C_MIN_RATE_MIN,
 				DRBD_C_MIN_RATE_MAX);
 		break;
+
+	case R_CONG_FILL:
+		m_strtoll_range(value, 's', name, DRBD_CONG_FILL_MIN,
+				DRBD_CONG_FILL_MAX);
+		break;
+
+	case R_CONG_EXTENTS:
+		m_strtoll_range(value, 1, name, DRBD_CONG_EXTENTS_MIN,
+				DRBD_CONG_EXTENTS_MAX);
+		break;
+
 	}
 }
 
@@ -281,6 +292,14 @@ static void derror(struct d_host_info *host, struct d_resource *res, char *text)
 		config_file, c_section_start, res->name, names_to_str(host->on_hosts), text);
 }
 
+void pdperror(char *text)
+{
+	config_valid = 0;
+	fprintf(stderr, "%s:%d: in proxy plugin section: %s.\n",
+		config_file, line, text);
+	exit(E_config_invalid);
+}
+
 static void pperror(struct d_host_info *host, struct d_proxy_info *proxy, char *text)
 {
 	config_valid = 0;
@@ -595,7 +614,7 @@ static struct d_option *parse_options_d(int token_switch, int token_option,
 					void *ctx)
 {
 	char *opt_name;
-	int token;
+	int token, token_group;
 	enum range_checks rc;
 
 	struct d_option *options = NULL, *ro = NULL;
@@ -603,11 +622,15 @@ static struct d_option *parse_options_d(int token_switch, int token_option,
 	fline = line;
 
 	while (1) {
-		token = yylex();
+		token_group = yylex();
+		/* Keep the higher bits in token_option, remove them from token. */
+		token = REMOVE_GROUP_FROM_TOKEN(token_group);
 		fline = line;
+
 		if (token == token_switch) {
 			options = APPEND(options, new_opt(yylval.txt, NULL));
-		} else if (token == token_option) {
+		} else if (token == token_option ||
+				GET_TOKEN_GROUP(token_option & token_group)) {
 			opt_name = yylval.txt;
 			check_and_change_deprecated_alias(&opt_name, token_option);
 			rc = yylval.rc;
@@ -615,7 +638,8 @@ static struct d_option *parse_options_d(int token_switch, int token_option,
 			range_check(rc, opt_name, yylval.txt);
 			ro = new_opt(opt_name, yylval.txt);
 			options = APPEND(options, ro);
-		} else if (token == token_delegate) {
+		} else if (token == token_delegate ||
+				GET_TOKEN_GROUP(token_delegate & token_group)) {
 			delegate(ctx);
 			continue;
 		} else if (token == TK_DEPRECATED_OPTION) {
@@ -1364,6 +1388,84 @@ void set_disk_in_res(struct d_resource *res)
 	}
 }
 
+void proxy_delegate(void *ctx)
+{
+	struct d_resource *res = (struct d_resource *)ctx;
+	int token;
+	struct d_option *options, *opt;
+	struct d_name *line, *word, **pnp;
+
+	opt = NULL;
+	token = yylex();
+	if (token != '{') {
+		fprintf(stderr,	"%s:%d: expected \"{\" after \"proxy\" keyword\n",
+				config_file, fline);
+		exit(E_config_invalid);
+	}
+
+	options = NULL;
+	while (1) {
+		pnp = &line;
+		while (1) {
+			token = yylex();
+			if (token == ';')
+				break;
+			if (token == '}') {
+				if (pnp == &line)
+					goto out;
+
+				fprintf(stderr,	"%s:%d: Missing \";\" before  \"}\"\n",
+					config_file, fline);
+				exit(E_config_invalid);
+			}
+
+			word = malloc(sizeof(struct d_name));
+			if (!word)
+				pdperror("out of memory.");
+			word->name = yylval.txt;
+			word->next = NULL;
+			*pnp = word;
+			pnp = &word->next;
+		}
+
+		opt = calloc(1, sizeof(struct d_option));
+		if (!opt)
+			pdperror("out of memory.");
+		opt->name = strdup(names_to_str(line));
+		options = APPEND(options, opt);
+		free_names(line);
+	}
+out:
+	res->proxy_plugins = options;
+}
+
+int parse_proxy_settings(struct d_resource *res, int flags)
+{
+	int token;
+
+	if (flags & PARSER_CHECK_PROXY_KEYWORD) {
+		token = yylex();
+		if (token != TK_PROXY) {
+			if (flags & PARSER_STOP_IF_INVALID) {
+				yyrestart(yyin); /* flushes flex's buffers */
+				return 1;
+			}
+
+			pe_expected_got("proxy", token);
+		}
+	}
+
+	EXP('{');
+
+	res->proxy_options =
+		parse_options_d(TK_PROXY_SWITCH,
+				TK_PROXY_OPTION | TK_PROXY_GROUP,
+				TK_PROXY_DELEGATE,
+				proxy_delegate, res);
+
+	return 0;
+}
+
 struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 {
 	struct d_resource* res;
@@ -1470,9 +1572,7 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 			break;
 		case TK_PROXY:
 			check_upr("proxy section", "%s:proxy", res->name);
-			EXP('{');
-			res->proxy_options =  parse_options(TK_PROXY_SWITCH,
-							    TK_PROXY_OPTION);
+			parse_proxy_settings(res, 0);
 			break;
 		case TK_DEVICE:
 			check_upr("device statement", "%s:device", res->name);
