@@ -155,7 +155,11 @@ struct option metaopt[] = {
 #define MD_BM_OFFSET_07        (MD_AL_OFFSET_07 + MD_AL_MAX_SECT_07)
 #define MD_RESERVED_SECT_07    ( (uint64_t)(128ULL << 11) )
 #define MD_BM_MAX_BYTE_07      ( (uint64_t)(MD_RESERVED_SECT_07 - MD_BM_OFFSET_07)*512 )
+#if BITS_PER_LONG == 32
 #define MD_BM_MAX_BYTE_FLEX    ( (uint64_t)(1ULL << (32-3)) )
+#else
+#define MD_BM_MAX_BYTE_FLEX    ( (uint64_t)(1ULL << (38-3)) )
+#endif
 
 #define DEFAULT_BM_BLOCK_SIZE  (1<<12)
 
@@ -216,6 +220,7 @@ struct md_cpu {
 	uint32_t flags;
 	uint64_t device_uuid;
 	uint32_t bm_bytes_per_bit;
+	uint32_t la_peer_max_bio_size;
 };
 
 /*
@@ -497,7 +502,8 @@ struct __packed md_on_disk_08 {
 	be_u32 al_nr_extents;	/* important for restoring the AL */
 	be_s32 bm_offset;	/* signed sector offset to the bitmap, from here */
 	be_u32 bm_bytes_per_bit;
-	be_u32 reserved_u32[4];
+	be_u32 la_peer_max_bio_size; /* last peer max_bio_size */
+	be_u32 reserved_u32[3];
 
 	char reserved[8 * 512 - (8*(UI_SIZE+3)+4*11)];
 };
@@ -518,6 +524,7 @@ void md_disk_08_to_cpu(struct md_cpu *cpu, const struct md_on_disk_08 *disk)
 	cpu->al_nr_extents = be32_to_cpu(disk->al_nr_extents.be);
 	cpu->bm_offset = be32_to_cpu(disk->bm_offset.be);
 	cpu->bm_bytes_per_bit = be32_to_cpu(disk->bm_bytes_per_bit.be);
+	cpu->la_peer_max_bio_size = be32_to_cpu(disk->la_peer_max_bio_size.be);
 }
 
 void md_cpu_to_disk_08(struct md_on_disk_08 *disk, const struct md_cpu *cpu)
@@ -535,6 +542,7 @@ void md_cpu_to_disk_08(struct md_on_disk_08 *disk, const struct md_cpu *cpu)
 	disk->al_nr_extents.be = cpu_to_be32(cpu->al_nr_extents);
 	disk->bm_offset.be = cpu_to_be32(cpu->bm_offset);
 	disk->bm_bytes_per_bit.be = cpu_to_be32(cpu->bm_bytes_per_bit);
+	disk->la_peer_max_bio_size.be = cpu_to_be32(cpu->la_peer_max_bio_size);
 	memset(disk->reserved, 0, sizeof(disk->reserved));
 }
 
@@ -1063,7 +1071,11 @@ void re_initialize_md_offsets(struct format *cfg)
 		md_size_sect = (md_size_sect + 7) & ~7ULL;         /* align on 4K blocks */
 
 		if (md_size_sect > (MD_BM_MAX_BYTE_FLEX>>9)) {
-			fprintf(stderr, "Device too large. We only support up to ~16TB.\n");
+			char ppbuf[10];
+			fprintf(stderr, "Device too large. We only support up to %s.\n",
+					ppsize(ppbuf, MD_BM_MAX_BYTE_FLEX << (3+2)));
+			if (BITS_PER_LONG == 32)
+				fprintf(stderr, "Maybe try a 64bit arch?\n");
 			exit(10);
 		}
 		/* plus the "drbd meta data super block",
@@ -1073,8 +1085,8 @@ void re_initialize_md_offsets(struct format *cfg)
 		cfg->md.bm_offset = -md_size_sect + MD_AL_OFFSET_07;
 		break;
 	}
-	cfg->al_offset = cfg->md_offset + cfg->md.al_offset * 512;
-	cfg->bm_offset = cfg->md_offset + cfg->md.bm_offset * 512;
+	cfg->al_offset = cfg->md_offset + cfg->md.al_offset * 512LL;
+	cfg->bm_offset = cfg->md_offset + cfg->md.bm_offset * 512LL;
 }
 
 /* MAYBE DOES DISK WRITES!! */
@@ -1387,8 +1399,8 @@ int v07_style_md_open(struct format *cfg)
 		return NO_VALID_MD_FOUND;
 	}
 
-	cfg->al_offset = cfg->md_offset + cfg->md.al_offset * 512;
-	cfg->bm_offset = cfg->md_offset + cfg->md.bm_offset * 512;
+	cfg->al_offset = cfg->md_offset + cfg->md.al_offset * 512LL;
+	cfg->bm_offset = cfg->md_offset + cfg->md.bm_offset * 512LL;
 
 	// For the case that someone modified la_sect by hand..
 	if( (cfg->md_index == DRBD_MD_INDEX_INTERNAL ||
@@ -1823,8 +1835,8 @@ int meta_dump_md(struct format *cfg, char **argv __attribute((unused)), int argc
 		cfg->md_offset = v07_style_md_get_byte_offset(
 			DRBD_MD_INDEX_FLEX_INT, cfg->lk_bd.bd_size);
 
-		cfg->al_offset = cfg->md_offset + cfg->md.al_offset * 512;
-		cfg->bm_offset = cfg->md_offset + cfg->md.bm_offset * 512;
+		cfg->al_offset = cfg->md_offset + cfg->md.al_offset * 512LL;
+		cfg->bm_offset = cfg->md_offset + cfg->md.bm_offset * 512LL;
 		cfg->bm_bytes = sizeof(long) *
 			bm_words(cfg->md.la_sect, cfg->md.bm_bytes_per_bit);
 	}
@@ -1857,6 +1869,8 @@ int meta_dump_md(struct format *cfg, char **argv __attribute((unused)), int argc
 			       cfg->md.bm_bytes_per_bit);
 			printf("device-uuid 0x"X64(016)";\n",
 			       cfg->md.device_uuid);
+			printf("la-peer-max-bio-size %d;\n",
+			       cfg->md.la_peer_max_bio_size);
 		}
 		printf("# bm-bytes %u;\n", cfg->bm_bytes);
 		printf_bm(cfg); /* pretty prints the whole bitmap */
@@ -2017,10 +2031,23 @@ int verify_dumpfile_or_restore(struct format *cfg, char **argv, int argc, int pa
 		cfg->md.bm_bytes_per_bit = yylval.u64;
 		EXP(TK_DEVICE_UUID); EXP(TK_U64); EXP(';');
 		cfg->md.device_uuid = yylval.u64;
+		int tok = yylex();
+		switch(tok) {
+		case TK_LA_BIO_SIZE:
+			EXP(TK_NUM); EXP(';');
+			cfg->md.la_peer_max_bio_size = yylval.u64;
+			break;
+		case TK_BM:
+			goto start_of_bm;
+		default:
+			md_parse_error(TK_BM, 0, "keyword 'bm' or 'la-peer-max-bio-size'");
+		}
 	} else {
 		cfg->md.bm_bytes_per_bit = 4096;
 	}
-	EXP(TK_BM); EXP('{');
+	EXP(TK_BM);
+start_of_bm:
+	EXP('{');
 	bm = (le_u64 *)on_disk_buffer;
 	i = 0;
 	bm_on_disk_off = cfg->bm_offset;
@@ -2765,7 +2792,7 @@ int v08_move_internal_md_after_resize(struct format *cfg)
 	 */
 
 	/* move activity log, fixed size immediately preceeding the "super block". */
-	cur_offset = old_offset + cfg->md.al_offset * 512;
+	cur_offset = old_offset + cfg->md.al_offset * 512LL;
 	PREAD(cfg->md_fd, on_disk_buffer, old_offset - cur_offset, cur_offset);
 	PWRITE(cfg->md_fd, on_disk_buffer, old_offset - cur_offset, cfg->al_offset);
 
@@ -2774,10 +2801,10 @@ int v08_move_internal_md_after_resize(struct format *cfg)
 	 * We do not initialize that part, we just leave "garbage" in there.
 	 * Once DRBD "agrees" on the new lower level device size, that part of
 	 * the bitmap will be handled by the module, anyways. */
-	old_bm_offset = old_offset + cfg->md.bm_offset * 512;
+	old_bm_offset = old_offset + cfg->md.bm_offset * 512LL;
 
 	/* move bitmap, in chunks, peel off from the end. */
-	cur_offset = old_offset + cfg->md.al_offset * 512 - buffer_size;
+	cur_offset = old_offset + cfg->md.al_offset * 512LL - buffer_size;
 	while (cur_offset > old_bm_offset) {
 		PREAD(cfg->md_fd, on_disk_buffer, buffer_size, cur_offset);
 		PWRITE(cfg->md_fd, on_disk_buffer, buffer_size,
@@ -3103,8 +3130,10 @@ int meta_chk_offline_resize(struct format *cfg, char **argv, int argc)
 	/* this is first, so that lk-bdev-info files are removed/updated
 	 * if we find valid meta data in the expected place. */
 	if (err == VALID_MD_FOUND) {
+		/* Do not clutter the output of the init script
 		printf("Found valid meta data in the expected location, %llu bytes into %s.\n",
-			(unsigned long long)cfg->md_offset, cfg->md_device_name);
+		       (unsigned long long)cfg->md_offset, cfg->md_device_name);
+		*/
 		/* create, delete or update the last known info */
 		err = lk_bdev_load(cfg->minor, &cfg->lk_bd);
 		if (cfg->md_index != DRBD_MD_INDEX_FLEX_INT)
